# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Random DNA generator."""

import random
import types
from typing import Any, Optional, Union

from pyglove.core import symbolic
from pyglove.core import typing as pg_typing
from pyglove.core.geno.base import DNA
from pyglove.core.geno.base import DNASpec
from pyglove.core.geno.dna_generator import DNAGenerator


@symbolic.members([
    ('seed', pg_typing.Int().noneable(), 'Random seed.')
])
class Random(DNAGenerator):
  """Random DNA generator."""

  def _setup(self):
    """Setup DNA spec."""
    if self.seed is None:
      self._random = random
    else:
      self._random = random.Random(self.seed)

  def _propose(self) -> DNA:
    """Propose a random DNA."""
    return random_dna(self._dna_spec, self._random)

  def _replay(self, trial_id: int, dna: DNA, reward: Any) -> None:
    """Replay the history to recover the last proposed DNA."""
    # NOTE(daiyip): If the seed is fixed, we want to reproduce the same
    # sequence of random examples, we can do this simply by repeating previous
    # generation process.
    if self.seed is not None:
      random_dna(self._dna_spec, self._random)


def random_dna(
    dna_spec: DNASpec,
    random_generator: Union[None, types.ModuleType, random.Random] = None,
    attach_spec: bool = True,
    previous_dna: Optional[DNA] = None
    ) -> DNA:
  """Generates a random DNA from a DNASpec.

  Example::

    spec = pg.geno.space([
        pg.geno.oneof([
            pg.geno.constant(),
            pg.geno.constant(),
            pg.geno.constant()
        ]),
        pg.geno.floatv(0.1, 0.2)
    ])

    print(pg.random_dna(spec))
    # DNA([2, 0.1123])

  Args:
    dna_spec: a DNASpec object.
    random_generator: a Python random generator.
    attach_spec: If True, attach the DNASpec to generated DNA.
    previous_dna: An optional DNA representing previous DNA. This field might
        be useful for generating stateful random DNAs.

  Returns:
    A DNA object.
  """
  return dna_spec.random_dna(
      random_generator or random, attach_spec, previous_dna)




# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.core.typing.callable_ext."""

import unittest
from pyglove.core.typing import annotation_conversion   # pylint: disable=unused-import
from pyglove.core.typing import callable_ext


class PresetArgValueTest(unittest.TestCase):
  """Tests for typing.PresetArgValue."""

  def test_basics(self):
    v = callable_ext.PresetArgValue()
    self.assertFalse(v.has_default)
    v = callable_ext.PresetArgValue(default=1)
    self.assertTrue(v.has_default)
    self.assertEqual(v.default, 1)
    self.assertEqual(repr(v), 'PresetArgValue(default=1)')
    self.assertEqual(str(v), 'PresetArgValue(\n  default=1\n)')
    self.assertEqual(
        callable_ext.PresetArgValue(), callable_ext.PresetArgValue()
    )
    self.assertEqual(
        callable_ext.PresetArgValue(1), callable_ext.PresetArgValue(1)
    )
    self.assertNotEqual(
        callable_ext.PresetArgValue(), callable_ext.PresetArgValue(default=1)
    )

  def test_inspect(self):

    def foo(
        x=callable_ext.PresetArgValue(),
        y=1,
        *,
        z=callable_ext.PresetArgValue(default=2)
    ):
      return x + y + z

    self.assertEqual(
        callable_ext.PresetArgValue.inspect(foo),
        dict(
            x=callable_ext.PresetArgValue(),
            z=callable_ext.PresetArgValue(default=2)
        )
    )

    def bar(x, y=1):
      return x + y

    self.assertEqual(
        callable_ext.PresetArgValue.inspect(bar),
        {}
    )

  def test_resolve_args(self):
    # Both positional and keyword arguments are from preset.
    self.assertEqual(
        callable_ext.PresetArgValue.resolve_args(
            call_args=[],
            call_kwargs=dict(),
            positional_arg_names=['x', 'y'],
            arg_defaults={
                'x': callable_ext.PresetArgValue(),
                'z': callable_ext.PresetArgValue(default=2),
            },
            preset_kwargs=dict(x=1, y=2, z=3, w=4)
        ),
        ([1], dict(z=3))
    )
    # Both positional and keyword arguments are absent and use the default.
    self.assertEqual(
        callable_ext.PresetArgValue.resolve_args(
            call_args=[],
            call_kwargs=dict(),
            positional_arg_names=['x', 'y'],
            arg_defaults={
                'x': callable_ext.PresetArgValue(default=1),
                'y': 0,
                'z': callable_ext.PresetArgValue(default=2),
            },
            preset_kwargs=dict()
        ),
        ([1, 0], dict(z=2))
    )
    # Positional args from preset, keyword argument use preset default.
    self.assertEqual(
        callable_ext.PresetArgValue.resolve_args(
            call_args=[],
            call_kwargs=dict(),
            positional_arg_names=['x', 'y'],
            arg_defaults={
                'x': callable_ext.PresetArgValue(),
                'y': 0,
                'z': callable_ext.PresetArgValue(default=2),
            },
            preset_kwargs=dict(x=1, y=2)
        ),
        ([1, 0], dict(z=2))
    )
    # Postional argument provided by user, which should take precedence over
    # the preset value.
    self.assertEqual(
        callable_ext.PresetArgValue.resolve_args(
            call_args=[2],
            call_kwargs={},
            positional_arg_names=['x', 'y'],
            arg_defaults={
                'x': callable_ext.PresetArgValue(),
                'y': 0,
                'z': callable_ext.PresetArgValue(default=2),
            },
            preset_kwargs=dict(x=1, y=2, z=3, w=4)
        ),
        ([2, 0], dict(z=3))
    )
    # Postional argument provided in keyword.
    self.assertEqual(
        callable_ext.PresetArgValue.resolve_args(
            call_args=[],
            call_kwargs=dict(x=2),
            positional_arg_names=['x', 'y'],
            arg_defaults={
                'x': callable_ext.PresetArgValue(),
                'y': 0,
                'z': callable_ext.PresetArgValue(default=2),
            },
            preset_kwargs=dict(x=1, y=2, z=3, w=4)
        ),
        ([2, 0], dict(z=3))
    )
    # Postional argument provided in keyword, and there are more args
    # (due to varargs)
    self.assertEqual(
        callable_ext.PresetArgValue.resolve_args(
            call_args=[1, 2, 3],
            call_kwargs=dict(x=2),
            positional_arg_names=['x', 'y'],
            arg_defaults={
                'x': callable_ext.PresetArgValue(),
                'y': 0,
                'z': callable_ext.PresetArgValue(default=2),
            },
            preset_kwargs=dict(x=1, y=2, z=3, w=4)
        ),
        ([1, 2, 3], dict(z=3))
    )
    # Required preset argument is not provided.
    with self.assertRaisesRegex(ValueError, 'Argument .* is not present.'):
      callable_ext.PresetArgValue.resolve_args(
          call_args=[],
          call_kwargs=dict(),
          positional_arg_names=['x'],
          arg_defaults={
              'x': callable_ext.PresetArgValue(),
          },
          preset_kwargs=dict()
      )

    # Include all preset kwargs.
    self.assertEqual(
        callable_ext.PresetArgValue.resolve_args(
            call_args=[],
            call_kwargs=dict(),
            positional_arg_names=['x', 'y'],
            arg_defaults={
                'x': callable_ext.PresetArgValue(),
                'y': 0,
                'z': callable_ext.PresetArgValue(default=2),
            },
            preset_kwargs=dict(x=1, y=2, z=3, w=4),
            include_all_preset_kwargs=True,
        ),
        ([1, 0], dict(z=3, w=4))
    )

  def test_preset_args(self):
    @callable_ext.enable_preset_args()
    def foo(
        x=callable_ext.PresetArgValue(),
        y=1,
        *args,
        z=callable_ext.PresetArgValue(default=2)
    ):
      del args
      return x + y + z

    with self.assertRaisesRegex(ValueError, 'Argument \'x\' is not present.'):
      foo()

    with callable_ext.preset_args(dict(x=1)):
      self.assertEqual(foo(), 1 + 1 + 2)

    # `y`` should not take precedence over the non-preset default.
    with callable_ext.preset_args(dict(x=1, y=2)):
      self.assertEqual(foo(), 1 + 1 + 2)

    with callable_ext.preset_args(dict(x=1, y=2, z=3)):
      self.assertEqual(foo(3), 3 + 1 + 3)

    with callable_ext.preset_args(dict(x=1, y=2, z=3)):
      self.assertEqual(foo(3, 3, z=4), 3 + 3 + 4)

  def test_enable_preset_args(self):

    # No side-effect if function does not have PresetArgValue.
    def bar(x, y):
      return x + y
    self.assertIs(bar, callable_ext.enable_preset_args()(bar))

    # `include_all_preset_kwargs` sets to False.
    @callable_ext.enable_preset_args()
    def baz(x, y=callable_ext.PresetArgValue(default=1), **kwargs):
      return x + y + sum(kwargs.values())

    with callable_ext.preset_args(dict(z=3, p=4)):
      self.assertEqual(baz(1), 1 + 1)

    # `include_all_prset_kwargs` is effective only when there is varkw.
    @callable_ext.enable_preset_args(include_all_preset_kwargs=True)
    def foo(x, y=callable_ext.PresetArgValue(default=1), **kwargs):
      return x + y + sum(kwargs.values())

    with callable_ext.preset_args(dict(z=3, p=4)):
      self.assertEqual(foo(1), 1 + 1 + 3 + 4)

    # `include_all_preset_kwargs` should be ignored if there is no varkw.
    @callable_ext.enable_preset_args(include_all_preset_kwargs=True)
    def fuz(x, y=callable_ext.PresetArgValue(default=1)):
      return x + y

    with callable_ext.preset_args(dict(y=2, z=3)):
      self.assertEqual(fuz(1), 1 + 2)

  def test_preset_args_nesting(self):
    @callable_ext.enable_preset_args()
    def foo(
        x=callable_ext.PresetArgValue(),
        y=1,
        *,
        z=callable_ext.PresetArgValue(default=2)
    ):
      return x + y + z

    def bar(inherit_preset: bool = False, **kwargs):
      with callable_ext.preset_args(
          {k: v + 1 for k, v in kwargs.items()},
          inherit_preset=inherit_preset
      ):
        return foo()

    with callable_ext.preset_args(dict(x=1)):
      self.assertEqual(foo(), 1 + 1 + 2)

      self.assertEqual(bar(x=1), 2 + 1 + 2)
      self.assertEqual(bar(x=1, z=2), 2 + 1 + 3)
      self.assertEqual(bar(x=1, z=3), 2 + 1 + 4)

      with self.assertRaisesRegex(ValueError, 'Argument \'x\' is not present.'):
        bar()
      self.assertEqual(bar(inherit_preset=True), 1 + 1 + 2)


class CallWithOptionalKeywordArgsTest(unittest.TestCase):
  """Tests for typing.CallWithOptionalKeywordArgs."""

  def test_function(self):
    """Test call with function."""

    def foo(a, b):
      return a + b

    f = callable_ext.CallableWithOptionalKeywordArgs(foo, ['b', 'c'])
    self.assertEqual(f(1, b=2, c=3), 3)

    def bar(a, **kwargs):
      return sum([a] + list(kwargs.values()))

    f = callable_ext.CallableWithOptionalKeywordArgs(bar, ['b', 'c'])
    self.assertEqual(f(1, b=2, c=3), 6)

  def testMethod(self):
    """Test call with method."""

    class A:

      def __call__(self, a, b):
        return a + b

    f = callable_ext.CallableWithOptionalKeywordArgs(A(), ['b', 'c'])
    self.assertEqual(f(1, b=2, c=3), 3)

    class B:

      def __call__(self, a, **kwargs):
        return sum([a] + list(kwargs.values()))

    f = callable_ext.CallableWithOptionalKeywordArgs(B(), ['b', 'c'])
    self.assertEqual(f(1, b=2, c=3), 6)

  def test_classmethod(self):
    """Test call with class method."""

    class A:

      @classmethod
      def foo(cls, a, b):
        return a + b

    f = callable_ext.CallableWithOptionalKeywordArgs(A.foo, ['b', 'c'])
    self.assertEqual(f(1, b=2, c=3), 3)


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python code execution."""

import ast
import contextlib
import io
import multiprocessing
import pickle
import queue
from typing import Any, Callable, Dict, Optional, Union

from pyglove.core import utils
from pyglove.core.coding import errors
from pyglove.core.coding import parsing
from pyglove.core.coding import permissions


# Key in returned dict that captures stdout.
STDOUT_KEY = '__stdout__'

# Key in the returned dict that represents the final result.
RESULT_KEY = '__result__'
_TLS_CODE_RUN_CONTEXT = '__code_run_context__'


@contextlib.contextmanager
def context(**kwargs):
  """Context manager to inject symbols for code execution."""
  ctx = get_context()
  ctx.update(kwargs)
  utils.thread_local_push(_TLS_CODE_RUN_CONTEXT, ctx)

  try:
    yield ctx
  finally:
    utils.thread_local_pop(_TLS_CODE_RUN_CONTEXT)


def get_context() -> Dict[str, Any]:
  """Gets the current context for code execution."""
  context_stack = utils.thread_local_get(_TLS_CODE_RUN_CONTEXT, None)
  return dict(context_stack[-1]) if context_stack else {}


def evaluate(
    code: str,
    *,
    global_vars: Optional[Dict[str, Any]] = None,
    permission: Optional[permissions.CodePermission] = None,
    returns_stdout: bool = False,
    outputs_intermediate: bool = False,
) -> Union[Any, Dict[str, Any]]:
  """Executes Python code.

  Features:
    * Fine-grained execution policy for limiting what APIs could be executed.
      This eliminates the need for sandboxing.
    * It exposes both the final results and intermediate results (variables).

  Args:
    code: Python code to run.
    global_vars: An optional dict as the globals that could be referenced by the
      code.
    permission: Permission for the Python code to run.
    returns_stdout: If True, the stdout (a str) will be returned.
    outputs_intermediate: Applicable when returns_stdout is False. If True,
      intermediate output will be outputted as a dict, with the last line's
      value accessible by key '__result__' and the std output accessible by
      key '__stdout__'. Otherwise the value of the last line will be returned.

  Returns:
    The value of the last line of the code block. Or a dict of variable
    names of all locals to their evaluated values as the output of the code to
    run. The value for the last line can be accessed by key '__result__'. Or the
    stdout as a str.
  """
  # Set up the permission and context.
  permission = permission or permissions.get_permission()
  ctx = dict(get_context())
  if global_vars:
    ctx.update(global_vars)

  # Parse the code str.
  code_block = parsing.parse(code, permission)
  global_vars, orig_global_vars = ctx, ctx.copy()

  # No code.
  if not code_block.body:   # pytype: disable=attribute-error
    return {} if outputs_intermediate else None

  stdout = io.StringIO()
  with contextlib.redirect_stdout(stdout):
    if hasattr(code_block.body[-1], 'value'):   # pytype: disable=attribute-error
      last_expr = code_block.body.pop()  # pytype: disable=attribute-error
      result_vars = [RESULT_KEY]

      if isinstance(last_expr, ast.Assign):
        for name_node in last_expr.targets:
          if isinstance(name_node, ast.Name):
            result_vars.append(name_node.id)

      last_expr = ast.Expression(last_expr.value)  # pytype: disable=attribute-error

      try:
        # Execute the lines before the last expression.
        # NOTE(daiyip): Only a `globals` dict is specified here, which will also
        # be used to output intermediate values by `exec`. We do not specify a
        # separate `locals` dict here, for - "If exec gets two separate objects
        # as globals and locals, the code will be executed as if it were
        # embedded in a class definition." - as the Python document explains.
        # The outcome is that new functions defined in the code block could not
        # be called by other newly defined functions.
        # Refer to https://stackoverflow.com/questions/
        # 73940751/why-cant-i-call-a-function-from-another-function-using-exec
        # for more details.
        exec(compile(code_block, '', mode='exec'), global_vars)  # pylint: disable=exec-used

        # Evaluate the last expression.
        result = eval(  # pylint: disable=eval-used
            compile(last_expr, '', mode='eval'), global_vars
        )
      except BaseException as e:
        raise errors.CodeError(code, e) from e

      for result_var in result_vars:
        global_vars[result_var] = result
    else:
      try:
        exec(compile(code_block, '', mode='exec'), global_vars)  # pylint: disable=exec-used
      except BaseException as e:
        raise errors.CodeError(code, e) from e
      global_vars[RESULT_KEY] = list(global_vars.values())[-1]

  if returns_stdout:
    return stdout.getvalue()
  if outputs_intermediate:
    outputs = {}
    for k, v in global_vars.items():
      if k == '__builtins__':
        continue
      if k not in orig_global_vars or v is not orig_global_vars[k]:
        outputs[k] = v
    # Add stdout to outputs.
    outputs[STDOUT_KEY] = stdout.getvalue()
    return outputs
  return global_vars[RESULT_KEY]


def sandbox_call(
    func: Callable[..., Any],
    *args,
    timeout: Optional[float] = None,
    **kwargs) -> Any:
  """Calls a function with sandboxing.

  Args:
    func: Function to call.
    *args: Positional arguments for `func`
    timeout: Execution timeout in seconds. If None, wait `func` to complete.
    **kwargs: Keyword arguments for `func`.

  Returns:
    Return value from `func`.

  Raises:
    TimeoutError: If the execution time exceeds the timeout.
    Exception: Exception raised from `func`.
  """
  def _call(q, *args, **kwargs):
    # NOTE(daiyip): if `q` is closed by the main process when `q.put` is called
    # on a subprocess, ValueError will be raised. This is okay since the main
    # process is no longer waiting for the result, and the subprocess could
    # recycled with non-zero error code, which does not affect the main
    # process.
    def _run():
      r = func(*args, **kwargs)
      try:
        return pickle.dumps(r)
      except BaseException as e:
        raise errors.SerializationError(
            f'Cannot serialize sandbox result: {r}', e
        ) from e

    try:
      q.put(_run())
    except Exception as e:  # pylint: disable=broad-exception-caught
      q.put(e)

  q = multiprocessing.Queue()
  p = multiprocessing.Process(
      target=_call, args=tuple([q] + list(args)), kwargs=kwargs
  )
  try:
    p.start()
    x = q.get(timeout=timeout)
  except queue.Empty as e:
    if p.is_alive():
      # We use `kill` instead of `terminate` to release process resources
      # right away.
      p.kill()
    raise TimeoutError(f'Execution time exceed {timeout} seconds.') from e
  finally:
    q.close()

  if isinstance(x, Exception):
    raise x
  try:
    return pickle.loads(x)
  except Exception as e:
    raise errors.SerializationError(
        'Cannot deserialize the output from sandbox.', e
    ) from e


def maybe_sandbox_call(
    func: Callable[..., Any],
    *args,
    sandbox: Optional[bool] = None,
    timeout: Optional[float] = None,
    **kwargs
) -> Any:
  """Maybe calls a function with sandboxing.

  Args:
    func: Function to call.
    *args: Postional args that will be passed to `func`.
    sandbox: If True, run code in sandbox; If False, run code in current
      process. If None, run in sandbox first, if the output could not be
      serialized and pass to current process, run the code again in current
      process.
    timeout: Execution timeout in seconds. If None, wait the code the complete.
    **kwargs: Keyword args that will be passed to `func`.

  Returns:
    The return value of `func`.

  Raises:
    TimeoutError: If the execution time exceeds the timeout.
    Exception: Exception  that are raised from `func`.
  """
  if sandbox is None:
    try:
      return sandbox_call(func, *args, timeout=timeout, **kwargs)
    # NOTE(daiyip): output could be serialized across processes, giving it
    # already finishes on sandbox, so it should be much safer to run under
    # current process.
    except errors.SerializationError:
      return func(*args, **kwargs)
  elif sandbox:
    return sandbox_call(func, *args, timeout=timeout, **kwargs)
  else:
    return func(*args, **kwargs)


def run(
    code: str,
    *,
    global_vars: Optional[Dict[str, Any]] = None,
    permission: Optional[permissions.CodePermission] = None,
    returns_stdout: bool = False,
    outputs_intermediate: bool = False,
    sandbox: Optional[bool] = None,
    timeout: Optional[float] = None,
) -> Union[Any, Dict[str, Any]]:
  """Executes Python code.

  Features:
    * Fine-grained execution policy for limiting what APIs could be executed.
      This eliminates the need for sandboxing.
    * It exposes both the final results and intermediate results (variables).

  Args:
    code: Python code to run.
    global_vars: An optional dict of
    permission: Permission for the Python code to run.
    returns_stdout: If True, the stdout (a str) will be returned.
    outputs_intermediate: Applicable when returns_stdout is False. If True,
      intermediate output will be outputted as a dict, with the last line's
      value accessible by key '__result__' and the std output accessible by
      key '__stdout__'. Otherwise the value of the last line will be returned.
    sandbox: If True, run code in sandbox; If False, run code in current
      process. If None, run in sandbox first, if the output could not be
      serialized and pass to current process, run the code again in current
      process.
    timeout: Execution timeout in seconds. If None, wait the code the complete.

  Returns:
    The value of the last line of the code block. Or a dict of variable
    names of all locals to their evaluated values as the output of the code to
    run. The value for the last line can be accessed by key '__result__'. Or the
    stdout as a str.

  Raises:
    TimeoutError: If the execution time exceeds the timeout.
    Exception: Exception  that are raised from the code.
  """
  return maybe_sandbox_call(
      evaluate, code=code, global_vars=global_vars, permission=permission,
      returns_stdout=returns_stdout, outputs_intermediate=outputs_intermediate,
      sandbox=sandbox, timeout=timeout
  )
